/*
 * Edge.hpp
 *
 * Definition of an arbitrary Edge
 *  Created on: Jul 5, 2011
 *      Author: kobitzsch
 */

#ifndef EDGE_HPP_
#define EDGE_HPP_

#include <iostream>
#include <sstream>
#include <string>
#include <algorithm>
#include <limits>

//Generates a Static Weighted edge
template< typename weight_type_slot, typename meta_weight_type_slot = std::numeric_limits<weight_type_slot> >
struct StaticUndirectedWeightData{
	/*******************************************/
	typedef weight_type_slot weight_type;
	typedef meta_weight_type_slot meta_weight;

	/*******************************************/
	StaticUndirectedWeightData( const weight_type & weight_ = 0 )
	: weight( weight_ )
	{}

	template< typename edge_with_additional_data_slot >
	StaticUndirectedWeightData( const edge_with_additional_data_slot & edge )
	: weight( edge.weight )
	{}

	//Cast the element to ints weight type
	inline operator const weight_type_slot &() const { return weight; }

	/*******************************************/
	inline void serialize( std::ostream & os ) const {
		os.write( reinterpret_cast<const char*> (& weight ), sizeof( weight ));
	}

	inline void deserialize( std::istream & is ) {
		is.read( reinterpret_cast<char*> (& weight ), sizeof( weight ));
	}

	bool operator==( const StaticUndirectedWeightData<weight_type_slot,meta_weight_type_slot> & other ) const {
		return other.weight == weight;
	}

	bool operator!=( const StaticUndirectedWeightData<weight_type_slot,meta_weight_type_slot> & other ) const {
		return other.weight != weight;
	}

	bool covers( const StaticUndirectedWeightData<weight_type_slot,meta_weight_type_slot> & other ) const {
		return weight <= other.weight;
	}

	std::string toString() const {
		std::ostringstream oss;
		oss << weight;
		return oss.str();
	}

	/*******************************************/
	const weight_type & getWeight() const { return weight; }
	weight_type & getWeight(){ return weight; }
	void setWeight( const weight_type & weight_ ){ weight = weight_; }

	/*******************************************/
	weight_type weight;
};

//Generates a Static Weighted edge
template< typename weight_type_slot, typename meta_weight_type_slot = std::numeric_limits<weight_type_slot> >
struct DirectedStaticWeightData : public StaticUndirectedWeightData< weight_type_slot, meta_weight_type_slot >{

	/*******************************************/
	DirectedStaticWeightData( const StaticUndirectedWeightData<weight_type_slot,meta_weight_type_slot> & data_without_direction )
	: StaticUndirectedWeightData< weight_type_slot, meta_weight_type_slot >( data_without_direction ),
	  forward( false ),
	  backward( false )
	{}

	DirectedStaticWeightData( const weight_type_slot & weight = 0, bool forward_ = false, bool backward_ = false )
	: StaticUndirectedWeightData< weight_type_slot, meta_weight_type_slot >( weight ),
	  forward( forward_ ),
	  backward( backward_ )
	{}

	template< typename edge_with_additional_data_slot >
	DirectedStaticWeightData( const edge_with_additional_data_slot & edge )
	: StaticUndirectedWeightData< weight_type_slot, meta_weight_type_slot >( edge.weight ),
	  forward( edge.forward ),
	  backward( edge.backward )
	{}

	/*******************************************/
	inline void serialize( std::ostream & os ) const {
		StaticUndirectedWeightData< weight_type_slot, meta_weight_type_slot >::serialize( os );
		os.write( reinterpret_cast<const char*> (& forward ), sizeof( forward ));
		os.write( reinterpret_cast<const char*> (& backward ), sizeof( backward ));
	}

	inline void deserialize( std::istream & is ) {
		StaticUndirectedWeightData< weight_type_slot, meta_weight_type_slot >::deserialize( is );
		is.read( reinterpret_cast<char*> (& forward ), sizeof( forward ));
		is.read( reinterpret_cast<char*> (& backward ), sizeof( backward ));
	}

	bool operator==( const DirectedStaticWeightData<weight_type_slot, meta_weight_type_slot> & other ) const {
		return other.forward == forward && other.backward == backward && StaticUndirectedWeightData< weight_type_slot, meta_weight_type_slot >::operator ==( other );
	}

	bool operator!=( const DirectedStaticWeightData<weight_type_slot, meta_weight_type_slot> & other ) const {
		return other.forward != forward && other.backward == backward && StaticUndirectedWeightData< weight_type_slot, meta_weight_type_slot >::operator !=( other );
	}

	bool covers( const DirectedStaticWeightData<weight_type_slot,meta_weight_type_slot> & other ) const {
		return (!other.forward || forward) && (!other.backward || backward) && StaticUndirectedWeightData< weight_type_slot, meta_weight_type_slot >::covers( other );
	}

	std::string toString() const {
		std::ostringstream oss;
		oss << (backward ? "<" : "-") << (forward ? ">" : "-") << " " << StaticUndirectedWeightData< weight_type_slot, meta_weight_type_slot >::toString();
		return oss.str();
	}

	/*******************************************/
	const bool & isForward() const { return forward; }
	const bool & isBackward() const { return backward; }
	void setForward( bool value ) { forward = value; }
	void setBackward( bool value ) { backward = value; }
	void reverse(){ std::swap( forward, backward ); }

	/*******************************************/
	bool forward;
	bool backward;
};

template< typename node_id_slot, typename edge_data_slot >
class Edge_t : public edge_data_slot{
public:
	/*******************************************/
	typedef edge_data_slot edge_data;

	/*******************************************/
	Edge_t()
	: edge_data_slot(),
	  target( (node_id_slot) 0 )
	{ /* do nothing */ }

	Edge_t( const node_id_slot & target_, const edge_data_slot & edge_data_ )
	: edge_data_slot(edge_data_),
	  target( target_ )
	{ /* do nothing */ }

	template< class edge_with_additional_information_type >
	Edge_t( const edge_with_additional_information_type & edge )
	: edge_data_slot( (typename edge_with_additional_information_type::edge_data) edge ),
	  target( edge.target )

	{ /* do nothing */ }

	/*******************************************/
	inline void serialize( std::ostream & os ) const {
		os.write( reinterpret_cast<const char*>(&target), sizeof( target ) );
		edge_data_slot::serialize( os );
	}

	inline void deserialize( std::istream & is ) {
		is.read( reinterpret_cast<char*>(&target), sizeof( target ) );
		edge_data_slot::deserialize( is );
	}

	bool operator==( const Edge_t<node_id_slot,edge_data_slot> & other ) const {
		return other.target == target && edge_data_slot::operator ==( other );
	}

	bool operator!=( const Edge_t<node_id_slot,edge_data_slot> & other ) const {
		return other.target != target && edge_data_slot::operator !=( other );
	}

	bool covers( const Edge_t<node_id_slot,edge_data_slot> & other ) const {
		return target == other.target && edge_data_slot::covers( other );
	}

	std::string toString() const {
		std::ostringstream oss;
		oss << target << " " << edge_data_slot::toString();
		return oss.str();
	}

	/*******************************************/
	const node_id_slot & getTarget() const { return target; }
	node_id_slot & getTarget() { return target; }
	void setTarget( const node_id_slot & target_ ){ target = target_; }

	/*******************************************/
	node_id_slot target;
};

//Definition of basic edge types
typedef Edge_t< unsigned int, DirectedStaticWeightData<unsigned int> > DirectedIntegerWeightedEdge;

#endif /* EDGE_HPP_ */
