/*
 * KStaticGraph.h
 *
 *  Created on: 05.08.2010
 *      Author: kobitzsch
 *
 * Template class for a simple graph, represented as an adjacency array
 */

#ifndef KBGRAPH_H_
#define KBGRAPH_H_

#include <vector>
#include <sstream>
#include <iostream>


		//Static graph
		template< typename NodeSlot, typename EdgeSlot >
		class KStaticGraph {

		public:
			typedef NodeSlot Node;
			typedef EdgeSlot Edge;

			typedef typename std::vector<Node>::iterator node_iterator;
			typedef typename std::vector<Node>::const_iterator const_node_iterator;
			typedef typename std::vector<Edge>::iterator edge_iterator;
			typedef typename std::vector<Edge>::const_iterator const_edge_iterator;

			KStaticGraph( unsigned int expected_number_of_nodes, unsigned int expected_number_of_edges )
				: constructed( false )
			{
				nodes.reserve( expected_number_of_nodes + 1);
				edges.reserve( expected_number_of_edges );
			}

			KStaticGraph()
				: constructed( false )
			{}

			unsigned int numberOfNodes() const { return nodes.size() - 1; }
			unsigned int numberOfEdges() const { return nodes[nodes.size()-1].firstEdgeID(); }

			node_iterator nodeBegin(){ return nodes.begin(); }
			node_iterator nodeEnd(){ return nodes.end(); }

			const_node_iterator nodeBegin() const { return nodes.begin(); }
			const_node_iterator nodeEnd() const { return nodes.end(); }

			/********************************************************************************
			 * Iterator based access
			 */
			edge_iterator edgeBeginITR( const unsigned int & id ){
				return edges.begin() + (unsigned int)nodes[(unsigned int)id].firstEdgeID();
			}

			edge_iterator edgeEndITR( const unsigned int & id ){
				return edges.begin() + (unsigned int)nodes[(unsigned int)id+1].firstEdgeID();
			}

			const_edge_iterator edgeBeginITR( const unsigned int & id ) const{
				return edges.begin() + (unsigned int)nodes[(unsigned int)id].firstEdgeID();
			}

			const_edge_iterator edgeEndITR( const unsigned int & id ) const{
				return edges.begin() + (unsigned int)nodes[(unsigned int)id+1].firstEdgeID();
			}

			/********************************************************************************
			 * ID based access
			 */
			unsigned int numberOfEdges( const unsigned int & id ) const {
				return nodes[ (unsigned int)id + 1 ].firstEdgeID() - nodes[ (unsigned int)id ].firstEdgeID();
			}

			const Node & getNode( const unsigned int & id ) const {
				return nodes[(unsigned int)id];
			}

			Node & getNode( const unsigned int & id ) {
				return nodes[(unsigned int)id];
			}

			const unsigned int & edgeBegin( const unsigned int & id ) const {
				return nodes[(unsigned int)id].firstEdgeID();
			}

			const unsigned int & edgeEnd( const unsigned int & id ) const {
				return nodes[(unsigned int)id+1].firstEdgeID();
			}

			const Edge & getEdge( const unsigned int & eid) const {
				return edges[ (unsigned int)eid ];
			}

			Edge & getEdge( const unsigned int & eid) {
				return edges[ (unsigned int)eid ];
			}

			unsigned int addNode(){
				nodes.push_back( Node( (unsigned int) edges.size() ) );
				return  nodes.size() - 1;
			}

			void addEdge( const unsigned int & id, const Edge & edge ){
				if( (unsigned int)id == nodes.size() - 1 ){
					edges.push_back( edge );
				} else {
					//This path is only added for service reasons... should not be used since it brings a considerable overhead
					std::cout << "Detected edge insert for a node unequal to the last one. This may be an error in the underlying algorithm." << std::endl;
					edges.insert( edges.begin() + nodes[ (unsigned int)id + 1 ].firstEdgeID(), edge );
					for( unsigned int i = (unsigned int)id + 1; i < nodes.size(); ++i )
						nodes[i].setFirstEdgeID( (unsigned int) (nodes[i].firstEdgeID() + 1) );
				}
			}

			void removeEdge( const unsigned int & id, const unsigned int & edge ){
				for( unsigned int eid = edge; (unsigned int)eid + 1< edges.size(); ++eid ){
					edges[eid] = edges[eid + 1];
				}
				for( unsigned int nid = ( id+1 ); (unsigned int)nid < nodes.size(); ++nid ){
					nodes[nid].setFirstEdgeID( (unsigned int) nodes[nid].firstEdgeID() - 1 );
				}
			}

			void finalizeConstruction(){
				constructed = true;
				nodes.push_back( Node( (unsigned int)edges.size() ) );
			}

			unsigned int byteSize() const {
				return Node::byteSize() * nodes.capacity() + Edge::byteSize() * edges.capacity();
			}

			//Load graph from file
			void deserialize( std::istream & is ){
				unsigned int num_nodes, num_edges;

				is.read( reinterpret_cast<char*>( & num_nodes ), sizeof( unsigned int ) );
				nodes.reserve( num_nodes );
				is.read( reinterpret_cast<char*>( & num_edges ), sizeof( unsigned int ) );
				edges.reserve( num_edges );

				std::cout << "[status] reading: " << num_nodes << " nodes, " << num_edges << " edges." << std::endl;

				for( unsigned int i = 0; i < num_nodes; ++i ){
					Node node;
					node.deserialize( is );
					nodes.push_back( node );
				}

				for( unsigned int i = 0; i < num_edges; ++i ){
					Edge edge;
					edge.deserialize( is );
					edges.push_back( edge );
				}
			}

			unsigned int locate( const unsigned int & id, const Edge & edge ) const {

				for( unsigned int eid = edgeBegin( id ); eid != edgeEnd( id ); ++eid ){
					if( edges[eid] == edge )
						return ( eid );
				}
				std::cout << "Edge: " << id << " :: " << edge << " not found." << std::endl;
				std::cout << "Edges of source: " << std::endl;
				for( unsigned int eid = nodes[ (unsigned int)id ].firstEdgeID(); eid != nodes[ (unsigned int)id + 1 ].firstEdgeID(); ++eid ){
					std::cout << "\t" << id << " :: " << edges[eid] << std::endl;
				}
				std::cout << "Edges at Target: " << std::endl;
				for( unsigned int eid = nodes[edge.getTarget()].firstEdgeID(); eid != nodes[edge.getTarget()+1].firstEdgeID(); ++eid ){
					std::cout << "\t" << edge.getTarget() << " :: " << edges[eid] << std::endl;
				}
				return edgeEnd(id);
			}

			unsigned int reverseEdgeID( const unsigned int & nid, const unsigned int & eid ) const{
				Edge edge = edges[eid];
				unsigned int from = edge.getTarget();
				edge.setTarget( nid );
				edge.reverse();
				return locate( from, edge );
			}

			std::string toString() const {
				std::ostringstream out_stream;
				out_stream << "Nodes: " << nodes.size() - 1 << "\n"
						   << "Edges: " << edges.size() << "\n";
				for( unsigned int i = 0; i < nodes.size(); ++i )
					out_stream << "#" << i << ": " << nodes[i].toString() << "\n";
				for( unsigned int i = 0; i < (unsigned int)nodes[nodes.size()-1].firstEdgeID(); ++i )
					out_stream << "#" << i << ": " << edges[i].toString() << "\n";
				return out_stream.str();
			}

			void reconstruct(){
				constructed = !constructed;
			}

			typename Edge::weight_type getEdgeCost( const unsigned int & from, const unsigned int & to ) const {
				return 0;
			}

			bool isSane() const {
				unsigned int num_nodes = numberOfNodes();
				for( unsigned int i = 0; i < num_nodes; ++i ){
					for( unsigned int eid = edgeBegin( (i) ); eid != edgeEnd( (i) ); ++eid ){
						if( reverseEdgeID( (i), eid ) == edgeEnd( getEdge( eid ).getTarget() ) )
							return false;
					}
				}
				return true;
			}


			//needs to be sorted for ch properties, if ch query desired
			void build( const std::vector< std::pair<unsigned int,Edge> > & edges ){
				unsigned int maxn = (unsigned int) 0;
				for( unsigned int i = 0; i < edges.size(); ++i ){
					maxn = std::max( edges[i].first, maxn );
					maxn = std::max( edges[i].second.getTarget(), maxn );
				}

				nodes.reserve( (unsigned int)maxn + 1 );
				this->edges.reserve( edges.size() );

				unsigned int pos = 0;
				for( unsigned int i = 0; i <= maxn; ++i ){
					addNode();
					while( edges[pos].first == (unsigned int) i ){
						addEdge( (unsigned int) i, edges[pos].second );
						++pos;
					}
				}
				finalizeConstruction();
			}

			void writeToDDSG( std::ostream & ofs ) const {
				ofs << "d\n";
				ofs << numberOfNodes() << " " << numberOfEdges()/2 << "\n";
				for( int i = 0; i < numberOfNodes(); ++i ){
					for( unsigned int eid = edgeBegin( (unsigned int) i ); eid != edgeEnd( (unsigned int) i ); ++eid ){
						const Edge & edge = getEdge( eid );
						int dir = 0;
						if( !edge.isForward() )
							dir = 2;
						else if( !edge.isBackward() )
							dir = 1;
						if( i < edge.getTarget() )
							ofs << i << " " << (unsigned int) edge.getTarget() << " " << edge.getWeight() << " " << dir << "\n";
					}
				}
			}

			void writeToDimacs( std::ostream & os ) const {
				os << "c Dimacs format created from KSG Graph\n";
				os << "p sp " << numberOfNodes() << " ";
				unsigned int num_edges = 0;
				for( unsigned int i = 0; i < numberOfEdges(); ++i )
					if( edges[i].isForward() ) ++num_edges;
				os << num_edges << "\n";
				for( int i = 0; i < numberOfNodes(); ++i ){
					for( unsigned int eid = edgeBegin( (unsigned int) i ); eid != edgeEnd( (unsigned int) i ); ++eid ){
						const Edge & edge = getEdge( eid );
						if( edge.isForward() ){
							os << "a " << i << " " << (unsigned int) edge.getTarget() << " " << edge.getWeight() << "\n";
						}
					}
				}
			}
		protected:
			bool constructed;
			std::vector< Node > nodes;
			std::vector< Edge > edges;

		};


#endif /* KBGRAPH_H_ */
