/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

#include <vector>
#include <tr1/tuple>
#include <iostream>
#include <string>
#include <map>
#include <sstream>

using namespace std;
using namespace std::tr1;

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy
  //typedef storage_slot Storage;
  typedef tuple<id_slot,key_slot> elementT;
  typedef vector<elementT> vectorT;
  typedef typename vectorT::iterator vectorTI;
  typedef typename vectorT::const_iterator vectorTCI;

  vectorT v;
  map<id_slot,unsigned int> pos;
  map<id_slot,bool> reached;

  void debug( string s ) const
  {
    cout << s << endl;
  }

  unsigned int findPos( id_slot k ) const
  {
    //unsigned int i;
    //for (i = 0; i < v.size(); i++) {
    //  if (get<0>(v.at(i)) == k) {
    //    return i;
    //  }
    //}
    //return v.size();

    return pos.find(k)->second;
  }

  unsigned int parent( unsigned int i )
  {
    return (i + 1) / 2 - 1;
  }

  unsigned int childl( unsigned int i )
  {
    return (i + 1) * 2 - 1;
  }

  unsigned int childr( unsigned int i )
  {
    return (i + 1) * 2;
  }

  void swap( unsigned int i, unsigned int j )
  {
    id_slot idI = get<0>(v.at(i));
    id_slot idJ = get<0>(v.at(j));

    //cerr << findPos(idI) << " " << findPos(idJ) << endl;
    unsigned int tmpPos = findPos(idI);
    pos.erase(idI);
    pos[idI] = findPos(idJ);
    pos.erase(idJ);
    pos[idJ] = tmpPos;
    //cerr << findPos(idI) << " " << findPos(idJ) << endl;

    elementT tmp = v.at(i);
    v.at(i) = v.at(j);
    v.at(j) = tmp;
  }

  void up_heap( unsigned int i )
  {
    //debug("up_heap()");
    while (i > 0) {
      if (get<1>(v.at(i)) > get<1>(v.at(parent(i)))) {
        //debug("returning");
        return;
      }

      //debug("swapping");
      swap(i, parent(i));
      i = parent(i);
    }
  }

  void down_heap( unsigned int i )
  {
    //debug("down_heap()");
    while (true) {
      unsigned int min = childl(i);

      if (min >= v.size()) {
        return;
      }

      if ((min + 1) < v.size() && get<1>(v.at(childr(i))) < get<1>(v.at(min))) {
        min = childr(i);
      }

      if (get<1>(v.at(i)) < get<1>(v.at(min))) {
        return;
      }

      swap(i, min);
      i = min;
    }
  }

  void decreaseKeyInt( unsigned int i, key_slot new_key ) {
    get<1>(v.at(i)) = new_key;
    up_heap(i);
  }

  void increaseKeyInt( unsigned int i, key_slot new_key ) {
    get<1>(v.at(i)) = new_key;
    down_heap(i);
  }

public:
  PriorityQueue( const GraphType & graph )
  : v()
  , pos()
  , reached()
  {
    debug("PriorityQueue()");
  }

  size_t size() const{
    debug(stringstream("size()=") << v.size());
    return v.size();
  }

  bool empty() const {
    debug(stringstream("empty()") << v.empty() ? "1" : "0");
    return v.empty();
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
    ostringstream ss;
    ss << ("push(") << id << "," << key << ")";
    debug(ss.str());
    //cout << id << "," << key;
    v.push_back(make_tuple(id,key));
    pos[id] = v.size() - 1;
    reached[id] = true;
    up_heap(v.size() - 1);
  }

  void deleteMin(){
    debug("deleteMin()");
    pos.erase(getMin());
    v.front() = v.back();
    pos.erase(get<0>(v.front()));
    pos[get<0>(v.front())] = 0;
    v.pop_back();
    down_heap(0);
    //cerr << "size " << v.size() << endl;
  }

  const id_slot & getMin() const{
    ostringstream ss;
    ss << ("getMin()=") << get<0>(v.front());
    debug(ss.str());
    //cout << get<0>(v.front()) << endl;
    return get<0>(v.front());
  }

  const key_slot & getMinKey() const {
    ostringstream ss;
    ss << ("getMinKey()=") << get<1>(v.front());
    debug(ss.str());
    //cout << get<1>(v.front()) << endl;
    return get<1>(v.front());
  }

  const key_slot & getCurrentKey( id_slot id ) const {
    ostringstream ss;
    ss << ("getCurrentKey()=") << get<1>(v.at(findPos(id)));
    debug(ss.str());

    //cerr << id << endl;
    //cerr << findPos(id) << endl;
    elementT e = v.at(findPos(id));
    //cout << id << "," << get<1>(e) << endl;

    return get<1>(e);
  }

  bool isReached( id_slot id ) const {
    debug(stringstream("isReached()=") << reached.count(id) != 0 ? "1" : "0");
    return reached.count(id) != 0;
  }

  bool contains( id_slot id ) const {
    ostringstream ss;
    ss << ("contains(") << id << ")=" << pos.count(id) != 0 ? "1" : "0";
    debug(ss.str());
    return pos.count(id) != 0;
  }

  void decreaseKey( id_slot id, key_slot new_key ) {
    ostringstream ss;
    ss << ("decreaseKey(") << id << "," << new_key << ")";
    debug(ss.str());

    unsigned int i = findPos(id);

    if (i == v.size()) {
      return;
    }

    decreaseKeyInt(i, new_key);
  }

  void increaseKey( id_slot id, key_slot new_key ){
    debug("increaseKey()");

    unsigned int i = findPos(id);

    if (i == v.size()) {
      return;
    }

    increaseKeyInt(i, new_key);
  }

  void updateKey( id_slot id, key_slot new_key ){
    debug("updateKey()");

    unsigned int i = findPos(id);

    if (i == v.size()) {
      return;
    }

    if (new_key > get<1>(v.at(i))) {
      increaseKeyInt(i, new_key);
    } else {
      decreaseKeyInt(i, new_key);
    }
  }

  void clear(){
    debug("clear()");
    v.clear();
    pos.clear();
    reached.clear();
  }
};


#endif /* BINARY_HEAP_HPP_ */
