/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

//#define DEBUG

#ifdef DEBUG
#define Debug(x) x
#else
#define Debug(x) 
#endif

#include <vector>
#include <iostream>
#include <string>
#include <cstring>
#include <list>
#include <tr1/unordered_map>
#include <assert.h>
#include <math.h>
#include <limits.h>

#define IH_SIZE 512

using namespace std;
using namespace std::tr1;

template <typename id_slot, typename key_slot>
struct IKElement {
  id_slot id;
  key_slot key;
};

template <typename id_slot, typename key_slot, typename meta_key_slot, int capacity>
class BinaryHeap {
  private:/*{{{*/
    typedef IKElement<id_slot, key_slot> Element;
    Element data[capacity + 2];
    int size; // current size
/*}}}*/
  public:
    BinaryHeap()/*{{{*/
    {
    }
/*}}}*/
    void clear()/*{{{*/
    {
      size = 0;

      data[0].key = meta_key_slot::min();

      // TODO: memcpy?
      for (int i = 1; i <= capacity + 1; i++) {
        data[i].key = meta_key_slot::max();
      }
    }
/*}}}*/
    id_slot & getMin() /*{{{*/
    {
      Debug(cout << "BH::getMin() = " << data[1].id << endl);

      return data[1].id;
    }
/*}}}*/
    key_slot & getMinKey() /*{{{*/
    {
      Debug(cout << "BH::getMinKey() = " << data[1].key << endl);

      return data[1].key;
    }
/*}}}*/
    void deleteMin()/*{{{*/
    {
      Debug(cout << "BH::deleteMin()" << endl);

      int hole = 1;
      int succ = 2;

      // move up min-path
      while (succ < size) {
        key_slot key1 = data[succ].key;
        key_slot key2 = data[succ + 1].key;

        if (key1 > key2) {
          succ++;
          data[hole].key = key2;
          data[hole].id = data[succ].id;
        } else {
          data[hole].key = key1;
          data[hole].id = data[succ].id;
        }

        hole = succ;
        succ <<= 1;
      }

      // bubble up rightmost element
      key_slot bubble = data[size].key;
      int pred = hole >> 1;

      // must terminate since infimum at root
      while (data[pred].key > bubble) {
        data[hole] = data[pred];
        hole = pred;
        pred >>= 1;
      }

      // move data to hole
      data[hole].key = bubble;
      data[hole].id = data[size].id;

      data[size].key = meta_key_slot::max();
      size -= 1;
    }
/*}}}*/
    void insert(id_slot id, key_slot key)/*{{{*/
    {
      Debug(cout << "BH::insert(" << id << "," << key << ")" << endl);
      size++;
      int hole = size;
      int pred = hole >> 1;
      key_slot predKey = data[pred].key;

      // terminates due to sentinel at 0
      while (predKey > key) {
        data[hole].key = predKey;
        data[hole].id = data[pred].id;
        hole = pred;
        pred >>= 1;
        predKey = data[pred].key;
      }

      data[hole].key = key;
      data[hole].id = id;
    }
/*}}}*/
    bool tryDecreaseKey(id_slot id, key_slot newKey)/*{{{*/
    {
      Debug(cout << "BH::decreaseKey(" << id << "," << newKey << ")" << endl);

      int hole = 1;

      while (hole <= size) {
        if (id == data[hole].id) {
          int pred = hole >> 1;

          // must terminate since infimum at root
          while (data[pred].key > newKey) {
            data[hole] = data[pred];
            hole = pred;
            pred >>= 1;
          }

          // reinsert element
          data[hole].key = newKey;
          data[hole].id = id;

          return true;
        }

        hole++;
      }

      return false;
    }
/*}}}*/
    void sortOutAndEmpty(Element *out)/*{{{*/
    {
      Debug(cout << "BH::sortOutAndEmpty(" << out << ")" << endl);

      Element* const outEnd = out + size;
      Element* const root = data + 1;

      while (out < outEnd) {
        // copy minimum
        *out = *root;
        out++;

        // bubble up second smallest as in deleteMin
        int hole = 1;
        int succ = 2;

        while (succ <= size) { // TODO: Warum <=, nicht wie oben < ?
          key_slot key1 = data[succ].key;
          key_slot key2 = data[succ + 1].key;

          if (key1 > key2) {
            succ++;
            data[hole].key = key2;
            data[hole].id = data[succ].id;
          } else {
            data[hole].key = key1;
            data[hole].id = data[succ].id;
          }

          hole = succ;
          succ <<= 1;
        }

        // mark hole as deleted
        data[hole].key = meta_key_slot::max();
      }

      size = 0;
    }
/*}}}*/
    int getSize() const/*{{{*/
    {
      Debug(cout << "BH::getSize() = " << size << endl);

      return size;
    }
/*}}}*/
};


//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  class Element {
    public:
      id_slot id;
      key_slot key;
      Element* next;
      Element* prev;

      Element(id_slot i = UINT_MAX, key_slot k = UINT_MAX)
      : id(i)
      , key(k)
      , next(NULL)
      , prev(NULL)
      {}
  };

  const unsigned int pSize;
  unsigned int maxWeight;
  unsigned int mask;

  id_slot removedId;
  id_slot removedKey;

  Element* b;
  Element* n;
  key_slot* p;
  BinaryHeap<id_slot, key_slot, meta_key_slot, IH_SIZE> insertHeap;

  unsigned int nCounter;
  unsigned int curPos;

  static void insertBefore(register Element* e, register Element* n) {
    register Element* nn = e->prev;
    n->next = e;
    n->prev = nn;
    nn->next = n;
    e->prev = n;
  }

  static void insertAfter(register Element* e, register Element* n) {
    register Element* nn = e->next;
    n->prev = e;
    n->next = nn;
    nn->prev = n;
    e->next = n;
  }

  static void cutOut(Element* e) {
    register Element* p = e->prev;
    register Element* n = e->next;
    n->prev = p;
    p->next = n;
  }

  void emptyInsertHeap() { // TODO: Write
#ifdef DEBUG
    cout << "emptyInsertHeap()" << endl;
#endif
    IKElement<id_slot, key_slot> *es = new IKElement<id_slot, key_slot>[IH_SIZE];
    insertHeap.sortOutAndEmpty(es);

    for (unsigned int i = 0; i < IH_SIZE; i++) {
      key_slot key = es[i].key;
      id_slot id = es[i].id;
      unsigned int pos = key & mask;
      if (b[pos].key != UINT_MAX) {
        Element* e = &n[nCounter++];
        e->id = id;
        e->key = key;
        insertAfter(&b[pos], e);
      } else {
        b[pos].id = id;
        b[pos].key = key;
      }
    }
  }

public:
  PriorityQueue( const GraphType & graph )
  : pSize(graph.numberOfNodes())
  , maxWeight(0)
  , insertHeap()
  {
    cout << "go" << endl;

    for (unsigned int eid = 0; eid < graph.numberOfEdges(); ++eid) {
      unsigned int newWeight = graph.getEdge(eid).getWeight();
      //unsigned int newWeight = 2000;

      if (newWeight > maxWeight) {
        maxWeight = newWeight;
      }
    }

    maxWeight++;
    maxWeight |= maxWeight >> 1;
    maxWeight |= maxWeight >> 2;
    maxWeight |= maxWeight >> 4;
    maxWeight |= maxWeight >> 8;
    maxWeight |= maxWeight >> 16;

    mask = maxWeight;

    maxWeight++;

    cout << "maxWeight: " << maxWeight << endl;

    p = new key_slot[pSize];
    n = new Element[pSize];
    b = new Element[maxWeight];
  }

  inline size_t size() const{
  }

  inline bool empty() const {
#ifdef DEBUG
    //cout << "size: " << size() << endl;
#endif
    // TODO: Implement efficiently
    return false;
    //return b[0].next == &b[0]; ?
    //return size() == 0;
  }

  void printList(Element* e) {
    Element *x = e->next;
    cout << "----------" << endl;
    cout << e << "(id:" << e->id << ",key:" << e->key << "," << e->prev << "," << e->next << endl;
    while (x != e) {
      cout << x << "(id:" << x->id << ",key:" << x->key << "," << x->prev << "," << x->next << endl;
      x = x->next;
    }
    cout << "----------" << endl;
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
#ifdef DEBUG/*{{{*/
    cout << "push(" << id << "," << key << ")" << endl;
#endif/*}}}*/
    if (insertHeap.getSize() == IH_SIZE) {
      emptyInsertHeap();
    }

    insertHeap.insert(id, key);
    p[id] = key;
  }

  inline const id_slot & getMin() {
#ifdef DEBUG/*{{{*/
    cout << "getMin()" << endl;
#endif/*}}}*/
    key_slot ihKey = insertHeap.getMinKey();
    unsigned int newCurPos = curPos;
    unsigned int endPos = ihKey & mask;

    while (newCurPos != endPos) {
      if (b[newCurPos].key != UINT_MAX) {
        removedKey = b[newCurPos].key; // TODO: fix?
        removedId = b[newCurPos].next->id;
        curPos = newCurPos;

        if (b[curPos].next == &b[curPos]) {
          b[curPos].id = UINT_MAX;
          b[curPos].key = UINT_MAX;
        } else {
          cutOut(b[curPos].next);
        }
        return removedId;
      }

      newCurPos = (newCurPos + 1) & mask;
    }

    removedKey = ihKey;
    removedId = insertHeap.getMin();
    insertHeap.deleteMin();

    return removedId;
  }

  inline const key_slot & getMinKey() const{
    return removedKey; // TODO: Problematic because reference changing?
  }

  void deleteMin(){
  }

  inline const key_slot & getCurrentKey( id_slot id ) {
#ifdef DEBUG/*{{{*/
    cout << "getCurrentKey(" << id << ") = " << p[id] << endl;
#endif/*}}}*/
    return p[id];
  }

  inline bool isReached( id_slot id ) const {
#ifdef DEBUG/*{{{*/
    cout << "isReached(" << id << ") = " << ((p[id] != UINT_MAX) ? "1" : "0") << endl;
#endif/*}}}*/
    return p[id] != UINT_MAX;
    //return p[id].key != UINT_MAX;
  }

  inline bool contains( id_slot id ) const {
  }

  void decreaseKey( id_slot id, key_slot new_key ) {
#ifdef DEBUG/*{{{*/
    cout << "decreaseKey(" << id << "," << new_key << ")" << endl;
#endif/*}}}*/

    if (insertHeap.tryDecreaseKey(id, new_key)) {
      p[id] = new_key;
      return;
    }

    unsigned int pos = new_key & mask;
    unsigned int oldPos = p[id] & mask;
    p[id] = new_key;
    Element* e = NULL;

    if (b[oldPos].id == id) {
      if (b[oldPos].next == &b[oldPos]) {
        b[oldPos].id = UINT_MAX;
        b[oldPos].key = UINT_MAX;
      } else {
        e = b[oldPos].next;
        e->next->prev = &b[oldPos];
        b[oldPos].id = e->id;
        b[oldPos].key = e->key;
        b[oldPos].next = e->next;
      }
    } else {
      e = b[oldPos].next;

      while (e->id != id) {
        e = e->next;
      }

      cutOut(e);
    }

    if (b[pos].key != UINT_MAX) {
      if (e == NULL) {
        e = &n[nCounter++];
      }
      e->id = id;
      e->key = new_key;
      insertAfter(&b[pos], e);
    } else {
      b[pos].id = id;
      b[pos].key = new_key;
    }
  }

  inline void increaseKey( id_slot id, key_slot new_key ) {
  }

  inline void updateKey( id_slot id, key_slot new_key ) {
  }

  inline void clear() {
    nCounter = 0;
    curPos = 0;

    insertHeap.clear();

    for (register unsigned int i = 0; i < maxWeight; i++) {
      //b[i] = new Element(UINT_MAX, UINT_MAX);
      b[i].id = UINT_MAX;
      b[i].key = UINT_MAX;
      b[i].next = &b[i];
      b[i].prev = &b[i];
    }

    //memset(p, UINT_MAX, pSize * sizeof(key_slot));
    for (unsigned int i = 0; i < pSize; i++) {
      p[i] = UINT_MAX;
    }
  }
};


#endif /* BINARY_HEAP_HPP_ */
