/*
 * BinaryHeap.hpp
 *
 *  Created on: Apr 26, 2011
 *      Author: kobitzsch
 */

#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

#include <vector>
#include <iostream>
#include <string>
#include <tr1/unordered_map>
#include <assert.h>

using namespace std;
using namespace std::tr1;

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  typedef std::pair<id_slot,key_slot> id_key;

  class Node {
    public:
      Node(id_key d)
      : prev(this)
      , next(this)
      , parent(NULL)
      , child(NULL)
      , data(d)
      {
      }

      Node* prev;
      Node* next;
      Node* parent;
      Node* child;
      id_key data;

      void setKey(key_slot k) {
        data.second = k;
      }

      static void add(Node* n, Node* root) {
        n->next = root->next;
        n->prev = root;

        root->next->prev = n;
        root->next = n;
      }

      static void merge(Node* n, Node* m) {
        m->prev->next = n->next;
        n->next->prev = m->prev;

        n->next = m;
        m->prev = n;
      }

      static void remove(Node* n) {
        n->next->prev = n->prev;
        n->prev->next = n->next;
      }
  };

  typedef Node* pointer;

  unordered_map<id_slot,pointer> _pointers;
  unordered_map<id_slot,bool> _reached;
  Node* _root;
  unsigned long _size;
  id_slot _null;

  void addRoot(Node* n) {
    if (_size++ == 0) {
      _root = n;
    } else {
      //cout << _root << endl;
      Node::add(n, _root);

      if (n->data.second < _root->data.second)
        _root = n;
    }
  }

  void addRoots(Node* n) {
    if (_size == 0) {
      _root = n;
    } else if (n->next == n){
      Node::add(n, _root);
    } else if (_root->next == _root) {
      Node::add(_root, n);
    } else {
      Node::merge(n, _root);
    }

    if (n->data.second < _root->data.second)
      _root = n;
  }

  void debug( string s ) const
  {
    //cout << s << endl;
  }

  unsigned int printTree(Node* root, string s = "") {
    unsigned int x = 1;
    cout << s << "root: " << root->data.first << "," << root->data.second << " (" << root << ", p: "<< root->parent << ", prev: " << root->prev << ")" << endl;
    if (root->child != NULL) {
      x += printTree(root->child, s + " ");
    }
    int i = 0;
    for (Node * n = root->next; n != root; n = n->next, i++) {
      cout << s << "next: " << n->data.first << "," << n->data.second << " (" << n << ", p: " << n->parent << ", prev: " << n->prev << ")" << endl;
      x += 1;
      if (n->child != NULL) {
        x += printTree(n->child, s + " ");
      }
    }

    return x;
  }

  void printRTree() {
    cout << "root: " << _root->data.first << "," << _root->data.second << endl;
    int i = 0;
    cout << _root->prev << endl;
    for (Node * n = _root->prev; n != _root && i <= 20; n = n->prev, i++) {
      cout << "prev: " << n->data.first << "," << n->data.second << endl;
    }
  }

public:
  PriorityQueue( const GraphType & graph )
  : _pointers()
  , _reached()
  , _root(NULL)
  , _size(0)
  , _null(0)
  {
    debug("PriorityQueue()");
  }

  size_t size() const{
    debug("size()");
    return _size;
  }

  bool empty() const {
    debug("empty()");
    return _size == 0;
  }

  //push an element onto the heap
  void push( const id_slot & id, const key_slot & key ){
    debug("push()");

    Node* n = new Node(make_pair(id, key));
    _pointers[id] = n;
    _reached[id] = true;
    addRoot(n);
  }

  void deleteMin(){
    debug("deleteMin()");
    Node* oldRoot = _root;
    //cout << "SIZE: " << printTree(_root) << ", size: " << _size << endl;;
    //cout << "-----------------------------------------" << endl;

    if (oldRoot->child != NULL) {
      oldRoot->child->parent = NULL;

      for (Node* n = oldRoot->child->next; n != oldRoot->child; n = n->next) {
        n->parent = NULL;
      }

      addRoots(oldRoot->child);
    }

    _root = oldRoot->prev;

    Node::remove(oldRoot);
    --_size;

    //printTree(_root);
    //cout << "-----------------------------------------" << endl;

    Node* n = _root;
    Node* p = _root;
    Node* pp = p->prev;
    Node* newMin = n;
    Node* newParent;
    Node* newChild;

    do {
      if (n == n->next) {
        break;
      }
      if (n->data.second <= n->next->data.second) {
        newParent = n;
        newChild = n->next;
      } else {
        newParent = n->next;
        newChild = n;
      }

      if (newParent->data.second < newMin->data.second) {
        newMin = newParent;
      }

      Node::remove(newChild);
      newChild->parent = newParent;

      if (newParent->child == NULL) {
        newParent->child = newChild;
        newChild->next = newChild;
        newChild->prev = newChild;
      } else {
        Node::add(newChild, newParent->child);

        if (newChild->data.second < newParent->child->data.second) {
          newParent->child = newChild;
        }
      }

      if (newChild->data.first == 1627550) {
        //cout << "BAM: " << newChild->next << "," << newChild->prev << "," << newChild->parent << endl;
      }

      n = newParent->next;
    } while (n != p && n != pp);

    if (newMin == oldRoot) {
      //cout << _size << endl;
      newMin = NULL;
    }
    _root = newMin;
    _pointers.erase(oldRoot->data.first);
    delete oldRoot;
    if (_root) {
      //cout << "SIZE: " << printTree(_root) << ", size: " << _size << endl;;
    }
  }
  //void deleteMin(){
  //  debug("deleteMin()");
  //  Node* oldRoot = _root;

  //  if (oldRoot->child != NULL) {
  //    oldRoot->child->parent = NULL;
  //    for (Node* n = oldRoot->child->next; n != oldRoot->child; n = n->next) {
  //      //cout << ">" << n->data.first << "," << n->data.second << endl;
  //      n->parent = NULL;
  //    }
  //    addRoots(oldRoot->child);
  //  }

  //  Node::remove(oldRoot);

  //  Node* p = _root->next;
  //  Node* pp = p->next;
  //  Node* n = _root->next;

  //  Node* newMin = n;
  //  Node* newChild;
  //  Node* newParent;

  //  do {
  //    cout << "newMin: " << newMin << endl;
  //    if (n == n->next)
  //      break;
  //    debug("0");
  //    //cout << n->data.second << "," << n->next->data.second << endl;
  //    if (n->data.second < n->next->data.second) {
  //      newParent = n;
  //      newChild = n->next;
  //    } else {
  //      newParent = n->next;
  //      newChild = n;
  //    }

  //    //cout << newParent->data.second << ",,," << newMin->data.second << endl;
  //    if (newParent->data.second < newMin->data.second && oldRoot != newParent)
  //      newMin = newParent;

  //    Node::remove(newChild);
  //    newChild->parent = newParent;

  //    if (newParent->child == NULL) {
  //      newChild->next = newChild;
  //      newChild->prev = newChild;
  //      newParent->child = newChild;
  //    } else {
  //      Node::add(newChild, newParent->child);

  //      if (newChild->data.second < newParent->child->data.second)
  //        newParent->child = newChild;
  //    }

  //    n = newParent->next;
  //  } while (n != p && n != pp);

  //  //cout << n->data.second << endl;
  //  //if (n->data.second < newMin->data.second) {
  //  //  newMin = n;
  //  //  n->child = NULL;
  //  //}

  //  debug("1");
  //  if (oldRoot == newMin) {
  //    debug("bam");
  //    cout << _size << "," << newMin << oldRoot << "," << oldRoot->next << endl;
  //    newMin = NULL;
  //  }

  //  _pointers.erase(oldRoot->data.first);
  //  delete oldRoot;
  //  _root = newMin;
  //  debug("2");
  //  if (_root != NULL) {
  //    //printTree(_root);
  //    //printRTree();
  //  }
  //  debug("3");
  //  --_size;
  //}

  const id_slot & getMin() {
    debug("getMin()");
    return _root->data.first;
  }

  const key_slot & getMinKey() {
    debug("getMinKey()");
    return _root->data.second;
  }

  const key_slot & getCurrentKey( id_slot id ) {
    debug("getCurrentKey()");
    //cout << id << endl;
    if (_pointers.count(id) == 0) {
      //cout << "fail" << endl;
      return _null;
    }
    return _pointers[id]->data.second;
  }

  bool isReached( id_slot id ) const {
    debug("isReached()");
    return _reached.count(id) != 0;
  }

  bool contains( id_slot id ) const {
    debug("contains()");
    return _pointers.count(id) != 0;
  }

  void decreaseKey( id_slot id, key_slot new_key ) {
    debug("decreaseKey()");
    //cout << id << "," << new_key << endl;

    Node* n = _pointers[id];
    debug("a");

    n->setKey(new_key);
    debug("b");

    if (n->parent != NULL) {
      debug("c");
      //cout << n->parent << endl;
      //cout << n->parent->child << endl;
      if (n->parent->child == n) {
        debug("d");
        if (n->next != n) {
          n->parent->child = n->next;
        } else {
          n->parent->child = NULL;
        }
      }

      debug("e");
      n->parent = NULL;
      debug("f");
      Node::remove(n);
      debug("g");
      Node::add(n, _root);
      debug("h");
    }

    debug("i");
    if (new_key < _root->data.second) {
      debug("j");
      _root = n;
    }
    debug("k");
  }

  void increaseKey( id_slot id, key_slot new_key ){
    debug("increaseKey()");
  }

  void updateKey( id_slot id, key_slot new_key ){
    debug("updateKey()");
  }

  void clear(){
    debug("clear()");
    _root = NULL;
    _size = 0;
    _pointers.clear();
    _reached.clear();
  }
};


#endif /* BINARY_HEAP_HPP_ */
