#ifndef BINARY_HEAP_HPP_
#define BINARY_HEAP_HPP_

//#define DEBUG

#ifdef DEBUG
#define Debug(x) x
#else
#define Debug(x) 
#endif

#include <vector>
#include <iostream>
#include <string>
#include <cstring>
#include <list>
#include <tr1/unordered_map>
#include <assert.h>
#include <math.h>
#include <limits.h>

using namespace std;
using namespace std::tr1;

const int KNBufferSize1 = 32;
const int KNN = 512;
const int KNKMAX = 64;
const int KNLevels = 4;
const int LogKNKMAX = 6;

template <typename id_slot, typename key_slot>
struct IKElement {
  id_slot id;
  key_slot key;
};

template <typename id_slot, typename key_slot, typename meta_key_slot, int capacity>
class BinaryHeap {
  private:/*{{{*/
    typedef IKElement<id_slot, key_slot> Element;
    Element data[capacity + 2];
    int size; // current size
/*}}}*/
  public:
    BinaryHeap()/*{{{*/
    {
    }
/*}}}*/
    void clear()/*{{{*/
    {
      size = 0;

      data[0].key = meta_key_slot::min();

      // TODO: memcpy?
      for (int i = 1; i <= capacity + 1; i++) {
        data[i].key = meta_key_slot::max();
      }
    }
/*}}}*/
    id_slot getMin() const/*{{{*/
    {
      Debug(cout << "BH::getMin() = " << data[1].id << endl);

      return data[1].id;
    }
/*}}}*/
    key_slot getMinKey() const/*{{{*/
    {
      Debug(cout << "BH::getMinKey() = " << data[1].key << endl);

      return data[1].key;
    }
/*}}}*/
    void deleteMin()/*{{{*/
    {
      Debug(cout << "BH::deleteMin()" << endl);

      int hole = 1;
      int succ = 2;

      // move up min-path
      while (succ < size) {
        key_slot key1 = data[succ].key;
        key_slot key2 = data[succ + 1].key;

        if (key1 > key2) {
          succ++;
          data[hole].key = key2;
          data[hole].id = data[succ].id;
        } else {
          data[hole].key = key1;
          data[hole].id = data[succ].id;
        }

        hole = succ;
        succ <<= 1;
      }

      // bubble up rightmost element
      key_slot bubble = data[size].key;
      int pred = hole >> 1;

      // must terminate since infimum at root
      while (data[pred].key > bubble) {
        data[hole] = data[pred];
        hole = pred;
        pred >>= 1;
      }

      // move data to hole
      data[hole].key = bubble;
      data[hole].id = data[size].id;

      data[size].key = meta_key_slot::max();
      size -= 1;
    }
/*}}}*/
    void insert(id_slot id, key_slot key)/*{{{*/
    {
      Debug(cout << "BH::insert(" << id << "," << key << ")" << endl);
      size++;
      int hole = size;
      int pred = hole >> 1;
      key_slot predKey = data[pred].key;

      // terminates due to sentinel at 0
      while (predKey > key) {
        data[hole].key = predKey;
        data[hole].id = data[pred].id;
        hole = pred;
        pred >>= 1;
        predKey = data[pred].key;
      }

      data[hole].key = key;
      data[hole].id = id;
    }
/*}}}*/
    // not to be called for elements not in binary heap
    void decreaseKey(id_slot id, key_slot newKey)/*{{{*/
    {
      Debug(cout << "BH::decreaseKey(" << id << "," << newKey << ")" << endl);

      int hole = 1;

      while (id != data[hole].id) {
        hole++;
      }

      int pred = hole >> 1;

      // must terminate since infimum at root
      while (data[pred].key > newKey) {
        data[hole] = data[pred];
        hole = pred;
        pred >>= 1;
      }

      // reinsert element
      data[hole].key = newKey;
      data[hole].id = id;
    }
/*}}}*/
    void sortOutAndEmpty(Element *out)/*{{{*/
    {
      Debug(cout << "BH::sortOutAndEmpty(" << out << ")" << endl);

      Element* const outEnd = out + size;
      Element* const root = data + 1;

      while (out < outEnd) {
        // copy minimum
        *out = *root;
        out++;

        // bubble up second smallest as in deleteMin
        int hole = 1;
        int succ = 2;

        while (succ <= size) { // TODO: Warum <=, nicht wie oben < ?
          key_slot key1 = data[succ].key;
          key_slot key2 = data[succ + 1].key;

          if (key1 > key2) {
            succ++;
            data[hole].key = key2;
            data[hole].id = data[succ].id;
          } else {
            data[hole].key = key1;
            data[hole].id = data[succ].id;
          }

          hole = succ;
          succ <<= 1;
        }

        // mark hole as deleted
        data[hole].key = meta_key_slot::max();
      }

      size = 0;
    }
/*}}}*/
    int getSize() const/*{{{*/
    {
      Debug(cout << "BH::getSize() = " << size << endl);

      return size;
    }
/*}}}*/
};

template <typename id_slot, typename key_slot, typename meta_key_slot, int capacity>
class SortedList {
  private:
    typedef IKElement<id_slot, key_slot> Element;
    Element data[capacity];
    int size;
    int startPos;

    void printList()
    {
      int pos = startPos;
      while (pos < startPos + size) {
        cout << "|" << data[pos].id << "," << data[pos].key;
        pos++;
      }
      cout << endl;
    }

  public:
    SortedList()
    {
    }

    void clear()
    {
      size = 0;
      startPos = 0;
    }

    id_slot getMin() const
    {
      return data[startPos].id;
    }

    key_slot getMinKey() const
    {
      return data[startPos].key;
    }

    void deleteMin()
    {
      //cout << "deleteMin" << endl;
      startPos++;
      size--;
      //printList();
    }

    void insert(id_slot id, key_slot key)
    {
      //cout << "insert: " << id << "," << key << endl;
      register int mid = startPos;
      while (data[mid].key <= key) {
        mid++;
      }
      //register int l = startPos;
      //register int r = startPos + size;
      //register int mid = (l + r) / 2;

      //while (l < r) {
      //  if (key < data[mid].key) {
      //    r = mid;
      //  } else {
      //    l = mid;
      //  }
      //  mid = (l + r) / 2;
      //}

      // TODO: use free front space

      register int pos = startPos + size;

      while (pos > mid) {
        data[pos].id = data[pos-1].id;
        data[pos].key = data[pos-1].key;

        pos--;
      }

      data[pos].id = id;
      data[pos].key = key;

      size++;

      if (startPos + size == capacity && startPos != 0) {
        for (int i = 0; i < size; i++) {
          data[i].key = data[i + startPos].key;
          data[i].id = data[i + startPos].id;
        }

        startPos = 0;
      }

      //cout << "startpos: " << startPos << ", size: " << size << endl;
      //printList();
    }

    void decreaseKey(id_slot id, key_slot newKey)
    {
      //cout << "decreaseKey" << endl;
      register int pos = startPos;
      while (data[pos].id != id) {
        pos++;
      }

      while (data[pos].key > newKey) {
        data[pos].id = data[pos-1].id;
        data[pos].key = data[pos-1].key;
      }

      data[pos].id = id;
      data[pos].key = newKey;

      //printList();
    }

    void sortOutAndEmpty(Element *out)
    {
      //memcpy(out, data+startPos, size);
      for (int i = startPos; i < size+startPos; i++) {
        out[i-startPos] = data[i];
      }
      clear();
    }

    int getSize() const
    {
      return size + startPos; // no more capacity
    }
};

template <typename id_slot, typename key_slot, typename meta_key_slot, int capacity>
class IKLoserTree {
  private:
    // Attributes/*{{{*/
    typedef IKElement<id_slot, key_slot> Element;

    struct Entry {
      key_slot key; // key of loser element (winner for 0)
      int index; // number of losing segment
    };

    // stack of empty segments
    int empty[capacity]; // indices of empty segments
    int lastFree; // where in "empty" is the last valid entry

    int size; // total number of elements stored
    int logK; // log of current tree size
    int k; // invariant k = 1 << logK

    Element dummy; // target of empty segment pointers

    // upper levels of loser trees
    // entry[0] contains the winner info
    Entry entry[capacity];

    // leaf information
    // indices 0..k-1
    Element* current[capacity]; // pointer to actual element
    Element* segment[capacity]; // start of segments
/*}}}*/
    int initWinner(int root)/*{{{*/
    {
      // Given any values in the leaves this routing recomputes upper levels of
      // the tree from scratch in linear time. Initialize entry[root].index and
      // the subtree rooted there, return winner index.

      if (root >= k) { // leaf reached
        return root - k; // TODO: negative values?
      } else {
        int left = initWinner(2 * root);
        int right = initWinner(2 * root + 1);
        key_slot lk = current[left]->key;
        key_slot rk = current[right]->key;

        if (lk <= rk) { // right subtree loses
          entry[root].index = right;
          entry[root].key = rk;
          return left;
        } else {
          entry[root].index = left;
          entry[root].key = lk;
          return right;
        }
      }
    }
/*}}}*/
    void updateOnInsert(int node, key_slot newKey,/*{{{*/
      int newIndex, key_slot* winnerKey,
      int* winnerIndex, // old winner
      int* mask) // distance from root
    {
      // First go up the tree all the way to the root, hand down old winner for
      // the respective subtree based on new value. Old winner and loser update
      // each node on the path to the root top down.

      if (node == 0) { // winner part of root
        *mask = 1 << (logK - 1);
        *winnerKey = entry[0].key;
        *winnerIndex = entry[0].index;

        if (newKey < entry[node].key) {
          entry[node].key = newKey;
          entry[node].index = newIndex;
        }
      } else {
        updateOnInsert(node >> 1, newKey, newIndex, winnerKey, winnerIndex, mask);
        key_slot loserKey = entry[node].key;
        int loserIndex = entry[node].index;

        if ((*winnerIndex & *mask) != (newIndex & *mask)) { // different subtrees
          if (newKey < loserKey) { // newKey will have influence here
            if (newKey < *winnerKey) { // old winner loses here
              entry[node].key = *winnerKey;
              entry[node].index = *winnerIndex;
            } else { // new entry loses here
              entry[node].key = newKey;
              entry[node].index = newIndex;
            }
          }

          *winnerKey = loserKey;
          *winnerIndex = loserIndex;
        }

        // nothing needs to be done if the winner came from the same subtree

        *mask >>= 1; // next level
      }
    }

    // free an empty segment
    void deallocateSegment(int index)
    {
      // reroute current pointer to some empty dummy segment with a sentinel
      // key
      current[index] = &dummy;

      // free memory
      delete [] segment[index];
      segment[index] = 0;

      // push on the stack of free segment indices
      lastFree++;
      empty[lastFree] = index;
    }
/*}}}*/
    void doubleK()/*{{{*/
    // make the tree two times as wide
    {
      // make all new entries empty and push them on the free stack
      for (int i = 2 * k - 1; i >= k; i--) {
        current[i] = &dummy;
        lastFree++;
        empty[lastFree] = i;
      }

      k *= 2;
      logK++;

      // recompute loser tree information
      rebuildLoserTree();
    }
/*}}}*/
    void compactTree()/*{{{*/
    // compact nonempty segments in the left half of the tree
    {
      int from = 0;
      int to = 0;

      // compact all nonempty segments to the left
      for (; from < k; from++) {
        if (current[from]->key != meta_key_slot::max()) {
          current[to] = current[from];
          segment[to] = segment[from];
          to++;
        }
      }

      // half degree as often as possible
      while (to < k / 2) {
        k /= 2;
        logK--;
      }

      // overwrite garbage and compact the stack of empty segments
      lastFree = -1; // none free

      for (; to < k; to++) {
        // push
        lastFree++;
        empty[lastFree] = to;
        current[to] = &dummy;
      }

      // recompute loser tree information
      rebuildLoserTree();
    }
/*}}}*/
    void rebuildLoserTree()/*{{{*/
    // rebuild loser tree information from the values in current
    {
      int winner = initWinner(1);
      entry[0].index = winner;
      entry[0].key = current[winner]->key;
    }
/*}}}*/
    bool segmentIsEmpty(int i)/*{{{*/
    {
      return (current[i]->key == meta_key_slot::max()) && (current[i] != &dummy);
    }
/*}}}*/
  public:
    IKLoserTree()/*{{{*/
    : lastFree(0)
    , size(0)
    , logK(0)
    , k(1)
    {
      dummy.key = meta_key_slot::max();
      empty[0] = 0;
      segment[0] = 0;
      current[0] = &dummy;

      rebuildLoserTree();
    }
/*}}}*/
    template <int LogK> void multiMergeUnrolled(Element* out, int l)/*{{{*/
    {
      Entry* currentPos;
      key_slot currentKey;
      int currentIndex; // leaf pointed to by current entry
      Element* outEnd = out + l;
      Entry* regEntry = entry;
      Element** regCurrent = current;
      int winnerIndex = regEntry[0].index;
      key_slot winnerKey= regEntry[0].key;
      Element* winnerPos;

      while (out < outEnd) {
        winnerPos = regCurrent[winnerIndex];

        // write result
        out->key = winnerKey;
        out->id = winnerPos->id;

        // advance winner segment
        winnerPos++;
        current[winnerIndex] = winnerPos;
        winnerKey = winnerPos->key;

        // remove winner segment if empty now
        if (winnerKey == meta_key_slot::max()) {
          deallocateSegment(winnerIndex);
        }

        out++;

        // update loser tree

#define TreeStep(L)\
      if ((1 << LogK) >= (1 << L)) {\
        Entry *pos##L = regEntry+((winnerIndex+(1<<LogK)) >> ((LogK-L)+1));\
        key_slot key##L = pos##L->key;\
        if (key##L < winnerKey) {\
          int index##L  = pos##L->index;\
          pos##L->key   = winnerKey;\
          pos##L->index = winnerIndex;\
          winnerKey     = key##L;\
          winnerIndex   = index##L;\
        }\
      }
        TreeStep(10);
        TreeStep(9);
        TreeStep(8);
        TreeStep(7);
        TreeStep(6);
        TreeStep(5);
        TreeStep(4);
        TreeStep(3);
        TreeStep(2);
        TreeStep(1);
      }

      regEntry[0].index = winnerIndex;
      regEntry[0].key = winnerKey;
    }
/*}}}*/
    void multiMerge(Element* out, int l)/*{{{*/
    // delete l smallest element to out
    {
      switch (logK) {
        case 0:
          memcpy(out, current[0], l * sizeof(Element));
          current[0] += l;
          entry[0].key = current[0]->key;

          if (segmentIsEmpty(0)) {
            deallocateSegment(0);
          }

          break;
        case 1:
          merge(current + 0, current + 1, out, l);
          rebuildLoserTree();

          if (segmentIsEmpty(0)) {
            deallocateSegment(0);
          }

          if (segmentIsEmpty(1)) {
            deallocateSegment(1);
          }
          break;
        case 2:
          merge4(current + 0, current + 1, current + 2, current + 3, out, l);
          rebuildLoserTree();

          if (segmentIsEmpty(0)) {
            deallocateSegment(0);
          }

          if (segmentIsEmpty(1)) {
            deallocateSegment(1);
          }
          if (segmentIsEmpty(2)) {
            deallocateSegment(2);
          }

          if (segmentIsEmpty(3)) {
            deallocateSegment(3);
          }
          break;
        case 3:
          multiMergeUnrolled<3>(out, l);
          break;
        case 4:
          multiMergeUnrolled<4>(out, l);
          break;
        case 5:
          multiMergeUnrolled<5>(out, l);
          break;
        case 6:
          multiMergeUnrolled<6>(out, l);
          break;
        case 7:
          multiMergeUnrolled<7>(out, l);
          break;
        case 8:
          multiMergeUnrolled<8>(out, l);
          break;
        case 9:
          multiMergeUnrolled<9>(out, l);
          break;
        case 10:
          multiMergeUnrolled<10>(out, l);
          break;
        default:
          multiMergeK(out, l);
          break;
      }

      size -= l;

      // compact tree if it got considerably smaller
      if (k > 1 && lastFree >= 3 * k / 5 - 1) {
        // using k/2 would be worst case inefficient
        compactTree();
      }
    }
/*}}}*/
    void multiMergeK(Element* out, int l)/*{{{*/
    {
      Entry* currentPos;
      key_slot currentKey;
      int currentIndex; // leaf pointed to by current entry
      int kReg = k;
      Element* outEnd = out + l;
      int winnerIndex = entry[0].index;
      key_slot winnerKey = entry[0].key;
      Element* winnerPos;

      while (out < outEnd) {
        winnerPos = current[winnerIndex];

        // write result
        out->key = winnerKey;
        out->id = winnerPos->id;

        // advance winner segment
        winnerPos++;
        current[winnerIndex] = winnerPos;
        winnerKey = winnerPos->key;

        // remove winner segment if empty now
        if (winnerKey == meta_key_slot::max()) {
          deallocateSegment(winnerIndex);
        }

        // go up the entry-tree
        for (int i = (winnerIndex + kReg) >> 1; i > 0; i >>= 1) {
          currentPos = entry + i;
          currentKey = currentPos->key;

          if (currentKey < winnerKey) {
            currentIndex = currentPos->index;
            currentPos->key = winnerKey;
            currentPos->index = winnerIndex;
            winnerKey = currentKey;
            winnerIndex = currentIndex;
          }
        }

        out++;
      }

      entry[0].index = winnerIndex;
      entry[0].key = winnerKey;
    }
/*}}}*/
    bool spaceIsAvailable()/*{{{*/
    {
      return (k < capacity) || (lastFree >= 0);
    }
/*}}}*/
    void insertSegment(Element* seg, int sz)/*{{{*/
    // insert segment beginng at seg
    // needs spaceIsAvailable() == 1
    {
      if (sz > 0) {
        // get a free slot
        if (lastFree < 0) { // tree is too small
          doubleK();
        }

        int index = empty[lastFree];
        lastFree--; // pop

        // link new segment
        current[index] = seg;
        segment[index] = seg;
        size += sz;

        // propagate new information up the tree
        key_slot dummyKey;
        int dummyIndex;
        int dummyMask;

        updateOnInsert((index + k) >> 1, seg->key, index,
                       &dummyKey, &dummyIndex, &dummyMask);
      } else {
        // immediately deallocate
        // this is not only an optimization but also needed to keep empty
        // segments from clogging up the tree
        delete [] seg;
      }
    }
/*}}}*/
    int getSize()/*{{{*/
    {
      return size;
    }
/*}}}*/
};

//id_slot : ID Type for stored elements
//key_slot: type of key_slot used
//meta key slot: information about stored data (e.g. std::numeric_limits<key>)
template< typename GraphType, typename id_slot, typename key_slot, typename meta_key_slot >
class PriorityQueue{
private:
  // Attributes/*{{{*/
  int _null;
  unsigned int n;

  typedef IKElement<id_slot, key_slot> Element;

  IKLoserTree<id_slot, key_slot, meta_key_slot, KNKMAX> tree[KNLevels];

  // one delete buffer for each tree (extra space for sentinel)
  Element buffer2[KNLevels][KNN + 1]; // tree->buffer2->buffer1
  Element* minBuffer2[KNLevels];

  // overall delete buffer
  Element buffer1[KNBufferSize1 + 1];
  Element* minBuffer1;

  // insert buffer
  BinaryHeap<key_slot, id_slot, meta_key_slot, KNN> insertHeap;

  // how many levels are active
  int activeLevels;

  // total size, not counting insertBuffer and buffer1
  int _size;

  unsigned int* _keys;
/*}}}*/
  PriorityQueue( const PriorityQueue & ){}  //do not copy
  void operator=( const PriorityQueue& ){}  //really, do not copy

  void refillBuffer1()/*{{{*/
  {
    int totalSize = 0;
    int sz;

    for (int i = activeLevels - 1; i >= 0; i--) {
      if ((&(buffer2[i][0]) + KNN) - minBuffer2[i] < KNBufferSize1) {
        sz = refillBuffer2(i);

        // max active level dry now?
        if (sz == 0 && i == activeLevels - 1) {
          activeLevels--;
        } else {
          totalSize += sz;
        }
      } else {
        totalSize += KNBufferSize1; // actually only a sufficient lower bound
      }
    }

    if (totalSize >= KNBufferSize1) { // buffer can be filled
      minBuffer1 = buffer1;
      sz = KNBufferSize1; // amount to be copied
      _size -= KNBufferSize1; // amount left in buffer2
    } else {
      minBuffer1 = buffer1 + KNBufferSize1 - totalSize;
      sz = totalSize;
      _size = 0;
    }

    // now call simplified refill routines
    // which can make the assumption that
    // they find all they are asked to find in the buffers
    minBuffer1 = buffer1 + KNBufferSize1 - sz;

    switch(activeLevels) {
      case 1: memcpy(minBuffer1, minBuffer2[0], sz * sizeof(Element));
              minBuffer2[0] += sz;
              break;
      case 2: merge(&(minBuffer2[0]), 
                    &(minBuffer2[1]), minBuffer1, sz);
              break;
      case 3: merge3(&(minBuffer2[0]), 
                     &(minBuffer2[1]),
                     &(minBuffer2[2]), minBuffer1, sz);
              break;
      case 4: merge4(&(minBuffer2[0]), 
                     &(minBuffer2[1]),
                     &(minBuffer2[2]),
                     &(minBuffer2[3]), minBuffer1, sz);
              break;
    }
  }
/*}}}*/
  int refillBuffer2(int j)/*{{{*/
  {
    Element *oldTarget;
    int deleteSize;
    int treeSize = tree[j].getSize();
    int bufferSize = (&(buffer2[j][0]) + KNN) - minBuffer2[j];

    if (treeSize + bufferSize >= KNN) { // buffer will be filled
      oldTarget = &(buffer2[j][0]);
      deleteSize = KNN - bufferSize;
    } else {
      oldTarget = &(buffer2[j][0]) + KNN - treeSize - bufferSize;
      deleteSize = treeSize;
    }

    // shift  rest to beginning
    // possible hack:
    // - use memcpy if no overlap
    memmove(oldTarget, minBuffer2[j], bufferSize * sizeof(Element));
    minBuffer2[j] = oldTarget;

    // fill remaining space from tree
    tree[j].multiMerge(oldTarget + bufferSize, deleteSize);
    return deleteSize + bufferSize;
  }
/*}}}*/
  int makeSpaceAvailable(int level)/*{{{*/
  {
    int finalLevel;

    if (level == activeLevels) {
      activeLevels++;
    }

    if (tree[level].spaceIsAvailable()) { 
      finalLevel = level;
    } else {
      finalLevel = makeSpaceAvailable(level + 1);

      int segmentSize = tree[level].getSize();
      Element *newSegment = new Element[segmentSize + 1];

      tree[level].multiMerge(newSegment, segmentSize); // empty this level
      newSegment[segmentSize].key = buffer1[KNBufferSize1].key; // sentinel

      // for queues where size << #inserts
      // it might make sense to stay in this level if
      // segmentSize < alpha * KNN * k^level for some alpha < 1
      tree[level + 1].insertSegment(newSegment, segmentSize);
    }

    return finalLevel;
  }
/*}}}*/
  void emptyInsertHeap()/*{{{*/
  {
    // build new segment
    Element *newSegment = new Element[KNN + 1];
    Element *newPos = newSegment;

    // put the new data there for now
    insertHeap.sortOutAndEmpty(newSegment);
    newSegment[KNN].key = meta_key_slot::max(); // sentinel

    // copy the buffer1 and buffer2[0] to temporary storage
    // (the temporary can be eliminated using some dirty tricks)
    const int tempSize = KNN + KNBufferSize1;
    Element temp[tempSize + 1];
    int sz1 = getSize1();
    int sz2 = getSize2(0);
    Element *pos = temp + tempSize - sz1 - sz2;
    memcpy(pos, minBuffer1, sz1 * sizeof(Element));
    memcpy(pos + sz1, minBuffer2[0], sz2 * sizeof(Element));
    temp[tempSize].key = meta_key_slot::max(); // sentinel

    // refill buffer1
    // (using more complicated code it could be made somewhat fuller
    // in certein circumstances)
    merge(&pos, &newPos, minBuffer1, sz1);

    // refill buffer2[0]
    // (as above we might want to take the opportunity
    // to make buffer2[0] fuller)
    merge(&pos, &newPos, minBuffer2[0], sz2);

    // merge the rest to the new segment
    // note that merge exactly trips into the footsteps
    // of itself
    merge(&pos, &newPos, newSegment, KNN);

    // and insert it
    int freeLevel = makeSpaceAvailable(0);
    tree[0].insertSegment(newSegment, KNN);

    // get rid of invalid level 2 buffers
    // by inserting them into tree 0 (which is almost empty in this case)
    if (freeLevel > 0) {
      for (int i = freeLevel;  i >= 0;  i--) { // reverse order not needed 
        // but would allow immediate refill
        newSegment = new Element[getSize2(i) + 1]; // with sentinel
        memcpy(newSegment, minBuffer2[i], (getSize2(i) + 1) * sizeof(Element));
        tree[0].insertSegment(newSegment, getSize2(i));
        minBuffer2[i] = buffer2[i] + KNN; // empty
      }
    }

    // update size
    _size += KNN; 

    // special case if the tree was empty before
    if (minBuffer1 == buffer1 + KNBufferSize1) {
      refillBuffer1();
    }
  }
/*}}}*/
  key_slot getSupremum() const/*{{{*/
  {
    return buffer2[0][KNN].key;
  }
/*}}}*/
  int getSize1() const/*{{{*/
  {
    return (buffer1 + KNBufferSize1) - minBuffer1;
  }
/*}}}*/
  int getSize2(int i) const/*{{{*/
  {
    return &(buffer2[i][KNN]) - minBuffer2[i];
  }
/*}}}*/
public:
  PriorityQueue( const GraphType & graph )/*{{{*/
  : _null(0)
  , insertHeap()
  , n(graph.numberOfNodes())
  {
    _keys = new unsigned int[n];
  }
/*}}}*/
  inline size_t size() const{/*{{{*/
    return _size + insertHeap.getSize() + ((buffer1 + KNBufferSize1) - minBuffer1);
  }
/*}}}*/
  inline bool empty() const {/*{{{*/
    // TODO: Implement efficiently
    // return size() == 0;
    return false;
  }
/*}}}*/
  void push( const id_slot & id, const key_slot & key ){/*{{{*/
    Debug(cout << "push(" << id << "," << key << ")" << endl);

    if (insertHeap.getSize() == KNN) {
      emptyInsertHeap();
    }

    insertHeap.insert(id, key);

    _keys[id] = key;
  }
/*}}}*/
  void deleteMin(){/*{{{*/
    Debug(cout << "deleteMin()" << endl);

    key_slot key1 = minBuffer1->key;
    key_slot key2 = insertHeap.getMinKey();

    if (key2 >= key1) {
      minBuffer1++;
      if (minBuffer1 == buffer1 + KNBufferSize1) {
        refillBuffer1();
      }
    } else {
      insertHeap.deleteMin();
    }
  }
/*}}}*/
  inline const id_slot getMin() {/*{{{*/
    key_slot key1 = minBuffer1->key;
    key_slot key2 = insertHeap.getMinKey();

    if (key2 >= key1) {
      if (key1 > _keys[minBuffer1->id]) {
        deleteMin();
        return getMin();
      } else {
        return minBuffer1->id;
      }
    } else {
      if (key2 > _keys[insertHeap.getMin()]) {
        deleteMin();
        return getMin();
      } else {
        return insertHeap.getMin();
      }
    }
  }
/*}}}*/
  inline const key_slot getMinKey() const{/*{{{*/
    key_slot key1 = minBuffer1->key;
    key_slot key2 = insertHeap.getMinKey();

    if (key2 >= key1) {
      return key1;
    } else {
      return key2;
    }
  }
/*}}}*/
  inline const key_slot & getCurrentKey( id_slot id ) {/*{{{*/
    Debug(cout << "getCurrentKey(" << id << ")" << endl);

    return _keys[id];
  }
/*}}}*/
  inline bool isReached( id_slot id ) const {/*{{{*/
    Debug(cout << "isReached(" << id << ")" << endl);

    return _keys[id] != 0;
  }
/*}}}*/
  void decreaseKey( id_slot id, key_slot new_key ) {
    Debug(cout << "decreaseKey(" << id << "," << new_key << ")" << endl);

    push(id, new_key);
  }

  inline void clear() {/*{{{*/
    _size = 0;
    activeLevels = 0;
    insertHeap.clear();

    buffer1[KNBufferSize1].key = meta_key_slot::max(); // sentinel
    minBuffer1 = buffer1 + KNBufferSize1; // empty

    for (int i = 0; i < KNLevels; i++) {
      buffer2[i][KNN].key = meta_key_slot::max(); // sentinel
      minBuffer2[i] = &(buffer2[i][KNN]); // empty
    }

    for (int i = 0; i <= n; i++) {
      _keys[i] = 0;
    }
/*}}}*/
  }

  // Unused methods/*{{{*/
  inline bool contains( id_slot id ) const {
  }

  inline void increaseKey( id_slot id, key_slot new_key ) {
  }

  inline void updateKey( id_slot id, key_slot new_key ) {
  }
/*}}}*/
};

// Auxiliary methods/*{{{*/
template <typename key_slot, typename id_slot>
void merge(IKElement<key_slot, id_slot> **f0,
           IKElement<key_slot, id_slot> **f1,
           IKElement<key_slot, id_slot>  *to, int sz) 
{
  IKElement<key_slot, id_slot> *from0   = *f0;
  IKElement<key_slot, id_slot> *from1   = *f1;
  IKElement<key_slot, id_slot> *done    = to + sz;
  key_slot      key0    = from0->key;
  key_slot      key1    = from1->key;

  while (to < done) {
    if (key1 <= key0) {
      to->key   = key1;
      to->id = from1->id; // note that this may be the same address
      from1++; // nach hinten schieben?
      key1 = from1->key;
    } else {
      to->key   = key0;
      to->id = from0->id; // note that this may be the same address
      from0++; // nach hinten schieben?
      key0 = from0->key;
    }
    to++;
  }
  *f0   = from0;
  *f1   = from1;
}


// merge sz element from the three sentinel terminated input
// sequences *f0, *f1 and *f2 to "to"
// advance *f0, *f1 and *f2 accordingly.
// require: at least sz nonsentinel elements available in f0, f1 and f2
// require: to may overwrite one of the sources as long as
//   *fx + sz is before the end of fx
template <typename key_slot, typename id_slot>
void merge3(IKElement<key_slot, id_slot> **f0,
           IKElement<key_slot, id_slot> **f1,
           IKElement<key_slot, id_slot> **f2,
           IKElement<key_slot, id_slot>  *to, int sz) 
{
  IKElement<key_slot, id_slot> *from0   = *f0;
  IKElement<key_slot, id_slot> *from1   = *f1;
  IKElement<key_slot, id_slot> *from2   = *f2;
  IKElement<key_slot, id_slot> *done    = to + sz;
  key_slot      key0    = from0->key;
  key_slot      key1    = from1->key;
  key_slot      key2    = from2->key;

  if (key0 < key1) {
    if (key1 < key2)   { goto s012; }
    else { 
      if (key2 < key0) { goto s201; }
      else             { goto s021; }
    }
  } else {
    if (key1 < key2) {
      if (key0 < key2) { goto s102; }
      else             { goto s120; }
    } else             { goto s210; }
  }

#define Merge3Case(a,b,c)\
  s ## a ## b ## c :\
  if (to == done) goto finish;\
  to->key = key ## a;\
  to->id = from ## a -> id;\
  to++;\
  from ## a ++;\
  key ## a = from ## a -> key;\
  if (key ## a < key ## b) goto s ## a ## b ## c;\
  if (key ## a < key ## c) goto s ## b ## a ## c;\
  goto s ## b ## c ## a;

  // the order is choosen in such a way that 
  // four of the trailing gotos can be eliminated by the optimizer
  Merge3Case(0, 1, 2);
  Merge3Case(1, 2, 0);
  Merge3Case(2, 0, 1);
  Merge3Case(1, 0, 2);
  Merge3Case(0, 2, 1);
  Merge3Case(2, 1, 0);

 finish:
  *f0   = from0;
  *f1   = from1;
  *f2   = from2;
}


// merge sz element from the three sentinel terminated input
// sequences *f0, *f1, *f2 and *f3 to "to"
// advance *f0, *f1, *f2 and *f3 accordingly.
// require: at least sz nonsentinel elements available in f0, f1, f2 and f2
// require: to may overwrite one of the sources as long as
//   *fx + sz is before the end of fx
template <typename key_slot, typename id_slot>
void merge4(IKElement<key_slot, id_slot> **f0,
           IKElement<key_slot, id_slot> **f1,
           IKElement<key_slot, id_slot> **f2,
           IKElement<key_slot, id_slot> **f3,
           IKElement<key_slot, id_slot>  *to, int sz) 
{
  IKElement<key_slot, id_slot> *from0   = *f0;
  IKElement<key_slot, id_slot> *from1   = *f1;
  IKElement<key_slot, id_slot> *from2   = *f2;
  IKElement<key_slot, id_slot> *from3   = *f3;
  IKElement<key_slot, id_slot> *done    = to + sz;
  key_slot      key0    = from0->key;
  key_slot      key1    = from1->key;
  key_slot      key2    = from2->key;
  key_slot      key3    = from3->key;

#define StartMerge4(a, b, c, d)\
  if (key##a <= key##b && key##b <= key##c && key##c <= key##d)\
    goto s ## a ## b ## c ## d;

  StartMerge4(0, 1, 2, 3);
  StartMerge4(1, 2, 3, 0);
  StartMerge4(2, 3, 0, 1);
  StartMerge4(3, 0, 1, 2);

  StartMerge4(0, 3, 1, 2);
  StartMerge4(3, 1, 2, 0);
  StartMerge4(1, 2, 0, 3);
  StartMerge4(2, 0, 3, 1);

  StartMerge4(0, 2, 3, 1);
  StartMerge4(2, 3, 1, 0);
  StartMerge4(3, 1, 0, 2);
  StartMerge4(1, 0, 2, 3);

  StartMerge4(2, 0, 1, 3);
  StartMerge4(0, 1, 3, 2);
  StartMerge4(1, 3, 2, 0);
  StartMerge4(3, 2, 0, 1);

  StartMerge4(3, 0, 2, 1);
  StartMerge4(0, 2, 1, 3);
  StartMerge4(2, 1, 3, 0);
  StartMerge4(1, 3, 0, 2);

  StartMerge4(1, 0, 3, 2);
  StartMerge4(0, 3, 2, 1);
  StartMerge4(3, 2, 1, 0);
  StartMerge4(2, 1, 0, 3);

#define Merge4Case(a, b, c, d)\
  s ## a ## b ## c ## d:\
  if (to == done) goto finish;\
  to->key = key ## a;\
  to->id = from ## a -> id;\
  to++;\
  from ## a ++;\
  key ## a = from ## a -> key;\
  if (key ## a < key ## c) {\
    if (key ## a < key ## b) { goto s ## a ## b ## c ## d; }\
    else                     { goto s ## b ## a ## c ## d; }\
  } else {\
    if (key ## a < key ## d) { goto s ## b ## c ## a ## d; }\
    else                     { goto s ## b ## c ## d ## a; }\
  }    
  
  Merge4Case(0, 1, 2, 3);
  Merge4Case(1, 2, 3, 0);
  Merge4Case(2, 3, 0, 1);
  Merge4Case(3, 0, 1, 2);

  Merge4Case(0, 3, 1, 2);
  Merge4Case(3, 1, 2, 0);
  Merge4Case(1, 2, 0, 3);
  Merge4Case(2, 0, 3, 1);

  Merge4Case(0, 2, 3, 1);
  Merge4Case(2, 3, 1, 0);
  Merge4Case(3, 1, 0, 2);
  Merge4Case(1, 0, 2, 3);

  Merge4Case(2, 0, 1, 3);
  Merge4Case(0, 1, 3, 2);
  Merge4Case(1, 3, 2, 0);
  Merge4Case(3, 2, 0, 1);

  Merge4Case(3, 0, 2, 1);
  Merge4Case(0, 2, 1, 3);
  Merge4Case(2, 1, 3, 0);
  Merge4Case(1, 3, 0, 2);

  Merge4Case(1, 0, 3, 2);
  Merge4Case(0, 3, 2, 1);
  Merge4Case(3, 2, 1, 0);
  Merge4Case(2, 1, 0, 3);

 finish:
  *f0   = from0;
  *f1   = from1;
  *f2   = from2;
  *f3   = from3;
}
/*}}}*/
#endif /* BINARY_HEAP_HPP_ */
